"""Allow user to edit and run a SPARQL query,
yield atoms sparql_vars/?, with ? the number of variables in the query.

"""


import os
try:
    from SPARQLWrapper import SPARQLWrapper, JSON
except ImportError:
    DISABLED = 'needs python module SPARQLWrapper'


NAME = 'SPARQL query'
TAGS = {'RDF', 'utils'}
INPUTS = {}
OUTPUTS = {'sparql_vars/*'}
DEFAULT_QUERY = 'data/sparql/solar_system.rq'
DEFAULT_ENDPOINT = 'https://query.wikidata.org/sparql'


def run_on(context:str, *,
           query:open=DEFAULT_QUERY, endpoint:str=DEFAULT_ENDPOINT,
           float_as_int:bool=True):
    """
    query -- (file containing) the full SPARQL query
    endpoint -- URL of the endpoint
    float_as_int -- convert float to integer by rounding them
    """
    if query and os.path.exists(query):
        with open(query) as fd:
            query = fd.read()
    yield from _atoms_from_query(endpoint, query, float_as_int=float_as_int)


def _atoms_from_query(endpoint, query, atom_name:str='sparql_vars', float_as_int:bool=False):
    """Wait for a valid database URI, and a SPARQL query.
    Yields all triplets returned by the query.
    The query need to yield three values, named object, relation and subject.
    """
    sparql = SPARQLWrapper(endpoint)
    sparql.setQuery(query)
    sparql.setReturnFormat(JSON)
    results = sparql.query()
    results = results.convert()
    variables = results['head']['vars']
    for result in results['results']['bindings']:
        yield '{}({}).\n'.format(atom_name, ','.join(_json_to_nuplet(result, variables, float_as_int=float_as_int)))


def _json_to_nuplet(json:dict, variables:tuple, float_as_int:bool=False) -> str:
    """Yield ASP arguments, created by parsing given json dict.

    Expected json object should be as {nuplet component: {'value': value}}.

    """
    values = {}  # variable: ASP value
    for variable in variables:
        yield _aspify_item(json[variable]['value'], float_as_int=float_as_int)


def _aspify_item(item, prefix='"', suffix='"', escapable_chars:str=set('"\\\''),
                 unwanted_chars=set('\n\r'), float_as_int:bool=False) -> str:
    """Return item, human readable"""
    if item.isnumeric() and int(item):
        return item
    elif float_as_int and item.count('.') == 1 and item.replace('.', '').isnumeric():
        return str(round(float(item)))
    else:  # it's a float, or a string
        return prefix + ''.join(
            ('\\' + c) if c in escapable_chars else c
            for c in item
            if c not in unwanted_chars
        ) + suffix



# encoding: utf8
"""Biseau script explaining the interface.

Comment/delete/modify `DISABLED = True`
in order to get it available from Biseau.

An easy way to test it is to add in GUI this rule:

    link(1,1).
    color(1,blue):- ichbineinberliner.
    color(1,green):- examplescriptwashere.

The node 1 will tell you if yes or no the scripts ran,
depending of the checkbox you checks.


"""

NAME = "Example script"
DISABLED = True  # comment/delete/modify this to enable this script in Biseau




# define the I/O of the scripts, if necessary based on the options.
#  note that the inputs/outputs function get the options, because they could
#  define the I/O. INPUTS/OUTPUTS attributes are not callable, just a set
#  defining the maximal I/O.
#  Here the input is constant whatever the options, but outputs depends.
def outputs(ich_bin_ein_berliner):
    return {'examplescriptwashere/0'} | ({'ichbineinberliner/0'} if ich_bin_ein_berliner else set())
INPUTS = {}
OUTPUTS = outputs(True)


def run_on(context:str, *,
           # Now begins the options of the script, shown in GUI with a widget.
           # Supported types: bool (checkbox), int/float (spinbox), str (text)
           # The default value gives the initial state of the widget in GUI.

           # here is a boolean
           ich_bin_ein_berliner:bool=True):


    # We do what we want with the context.
    context += '\nexamplescriptwashere.'

    # Which will probably depend of the options.
    if ich_bin_ein_berliner:
        context += '\nichbineinberliner.'

    # But the important part is to return it.
    return context



"""Core functions implementing the main behaviors.

Call example in main.

"""
import os
import io
import time
import json
import tempfile
from PIL import Image
from collections import OrderedDict
from . import utils
from . import Script
from . import asp_to_dot
from . import dot_writer
from . import module_loader
from biseau.script import solve_context, Script


EXT_TO_TYPE = utils.reverse_dict({
    'Python': {'.py'},
    'ASP': {'.lp'},
    'json/ASP': {'.json'},
}, multiple_values=True, aggregator=lambda x: next(iter(x)))
LOADABLE = {'Python', 'ASP', 'json/ASP'}


def single_image_from_filenames(fnames:[str], outfile:str=None, dotfile:str=None, nb_model:int=0, return_image:bool=True, dot_prog:str='dot', verbosity:int=0) -> Image or None:
    pipeline = build_pipeline(fnames, verbosity)
    final_context = run(pipeline, verbosity=verbosity)
    return compile_to_single_image(final_context, nb_model=nb_model, outfile=outfile, dotfile=dotfile, return_image=return_image, dot_prog=dot_prog, verbosity=verbosity)

def multiple_images_from_filenames(fnames:[str], outfile:str='out-{}.png', dotfile:str=None, nb_model:int=0, return_image:bool=True, dot_prog:str='dot', verbosity:int=0) -> Image or None:
    pipeline = build_pipeline(fnames, verbosity)
    final_context = run(pipeline, verbosity=verbosity)
    return tuple(compile_to_images(final_context, nb_model=nb_model, outfile_template=outfile, dotfile_template=dotfile, return_image=return_image, dot_prog=dot_prog, verbosity=verbosity))

def gif_from_filenames(fnames:[str], giffile:str=None, dotfile_template:str=None, duration:int=1000, loop:int=0, nb_model:int=0, dot_prog:str='dot', verbosity:int=0) -> bytes or str:
    """Make a gif, with each ASP model as an image. Save it in outfile and dotfile_template"""
    pipeline = build_pipeline(fnames, verbosity)
    final_context = run(pipeline, verbosity=verbosity)
    first, *lasts = compile_to_images(final_context, dotfile_template=dotfile_template, return_image=True, nb_model=nb_model, dot_prog=dot_prog, verbosity=verbosity)
    output = io.BytesIO() if giffile is None else giffile
    first.save(output, format='gif', append_images=lasts, duration=duration, loop=loop, save_all=True)
    return output.getvalue() if giffile is None else giffile


def build_pipeline(fnames:[str], verbosity:int=0) -> [Script]:
    "Yield scripts found in given filenames"
    for fname in fnames:
        if isinstance(fname, Script):
            yield fname
            continue
        ext = os.path.splitext(fname)[1]
        ftype = EXT_TO_TYPE.get(ext, 'unknow type')
        if ftype not in LOADABLE:
            raise ValueError(f"The type '{ftype}' (extension {ext}) can't be loaded")
        yield from module_loader.build_scripts_from_file(fname)

def build_pipeline_from_configfile(config:str, verbosity:int=0) -> [Script]:
    with open(config) as fd:
        configdata = json.loads(fd.read(), object_pairs_hook=OrderedDict)
    yield from build_pipeline_from_json(configdata)

def build_pipeline_from_json(jsondata:dict, verbosity:int=0) -> [Script]:
    if isinstance(jsondata, list):
        for item in jsondata:
            yield from build_pipeline_from_json(item)
    for name, options in jsondata.items():
        scripts = module_loader.build_scripts_from_file(name, options)
        for script in scripts:
            if isinstance(options, dict):
                script.options_values.update(options)
            yield script

build_pipeline.from_configfile = build_pipeline_from_configfile
build_pipeline.from_json = build_pipeline_from_json


def yield_run(scripts:[Script], initial_context:str='', verbosity:int=0) -> (str, float):
    "Yield context and duration found at each step of the running process"
    if verbosity >= 1:
        scripts = tuple(scripts)
        print(f"RUNNING {len(scripts)} SCRIPTS…")
    run_start = time.time()
    context = initial_context
    for idx, script in enumerate(scripts, start=1):
        script_start = time.time()
        context = Script.call(script, context, verbosity)
        script_time = round(time.time() - script_start, 2)
        if verbosity >= 1:
            print(f"SCRIPT {idx}: {script.name} added {len(context.splitlines())} lines to the context in {script_time}s.")
        yield context, script_time
    run_time = round(time.time() - run_start, 2)
    if verbosity >= 1:
        print(f"RUN {len(context.splitlines())} lines built in {run_time}s.")


def run(scripts:[Script], initial_context:str='', verbosity:int=0) -> str:
    "Return the final context obtained by running given pipeline on initial_context"
    out = None
    for out, _ in yield_run(scripts, initial_context, verbosity):
        pass
    return out


def compile_to_single_image(context:str, outfile:str=None, dotfile:str=None,
                            nb_model:int=0, return_image:bool=True,
                            dot_prog:str='dot', verbosity:int=0) -> Image or None:
    "Return a pillow.Image object, or write it to outfile if given"
    configs = asp_to_dot.visual_config_from_asp(
        solve_context(context, nb_model=nb_model)
    )
    dot = dot_writer.one_graph_from_configs(configs)
    del_outfile = False
    if outfile is None:
        with tempfile.NamedTemporaryFile(delete=False) as fd:
            outfile = fd.name
        del_outfile = True
    dot = dot_writer.dot_to_png(dot, outfile, dotfile=dotfile, prog=dot_prog)
    if return_image:
        img = Image.open(outfile)
        if del_outfile:
            os.unlink(outfile)
        return img


def compile_to_images(context:str, outfile_template:str=None, dotfile_template:str=None,
                      nb_model:int=0, return_image:bool=True,
                      dot_prog:str='dot', verbosity:int=0) -> [Image]:
    """Yield pillow.Image objects, and write it to outfile if given

    outfile_template -- template name for png files, or None
    dotfile_template -- template name for dot files, or None

    """
    if outfile_template and '{}' not in outfile_template:
        raise ValueError("Outfile argument is not a valid template")
    if dotfile_template and '{}' not in dotfile_template:
        raise ValueError("Dotfile argument is not a valid template")
    dots = compile_context_to_dots(context, nb_model)
    for idx, dot in enumerate(dots, start=1):
        del_outfile = False
        if outfile_template is None:
            with tempfile.NamedTemporaryFile(delete=False) as fd:
                outfile = fd.name
            del_outfile = True
        else:
            outfile = outfile_template.format(idx)
        dotfile = dotfile_template.format(idx) if dotfile_template else None
        dot = dot_writer.dot_to_png(dot, outfile, dotfile=dotfile, prog=dot_prog)
        if return_image:
            img = Image.open(outfile)
            if del_outfile:
                os.unlink(outfile)
            yield img


def compile_context_to_dots(context:str, nb_model:int=0) -> [str]:
    "Yield a dot string for each found model"
    configs = asp_to_dot.visual_config_from_asp(
        solve_context(context, nb_model=nb_model)
    )
    yield from dot_writer.multiple_graphs_from_configs(configs)


def compile_context_to_dot(context:str, nb_model:int=0) -> str:
    "Return a dot string for all found model"
    configs = asp_to_dot.visual_config_from_asp(
        solve_context(context, nb_model=nb_model)
    )
    return '\n'.join(dot_writer.one_graph_from_configs(configs))



# encoding: utf8
"""Various helpers"""

import os
import inspect
from functools import partial, wraps
from collections import defaultdict


HANDLED_COLORS = {'red', 'green', 'blue'}


def color_from_colors(colors:set) -> str:
    """Return one color that represents the given ones"""
    if all(color in HANDLED_COLORS for color in colors):

        if len(colors) == 1: return next(iter(colors))
        if len(colors) == 2:
            if 'red' in colors:
                if 'blue' in colors: return 'magenta'
                else: return 'yellow'  # red and green
            else: return 'cyan'  # blue and green
        if len(colors) == 3: return 'white'
    elif len(colors) == 1:  # special case where only one non handled color is given
        return next(iter(colors))
    raise ValueError('UNVALID COLORS: ' + ', '.join(colors))


def normalized_path(path:str) -> str:
    return os.path.expanduser(os.path.abspath(os.path.expanduser(path)))


def join_on_genstr(funcgen:callable, joiner:str='') -> str:
    """If funcgen is a generator, will join its result on call

    >>> join_on_genstr(lambda: 'abc')()
    'abc'
    >>> def g(): yield from 'abc'
    >>> join_on_genstr(g)()
    'abc'
    >>> def g(c): return c+'bc'
    >>> join_on_genstr(g)('a')
    'abc'

    """
    if inspect.isgeneratorfunction(funcgen) or inspect.isgenerator(funcgen):
        return wraps(funcgen)(lambda *a, **k: joiner.join(funcgen(*a, **k)))
    return funcgen


def compile_python_code(code:str) -> dict:
    return compile(code, '<string>', 'exec')
def run_compiled_python_code(code:'code', namespace:dict=None) -> dict:
    namespace = namespace or {}
    exec(code, namespace)
    return namespace
def run_python_code(code:str, namespace:dict=None) -> dict:
    """

    >>> run_python_code('a = 1')['a']
    1
    >>> b = 1
    >>> run_python_code('b = 2')['b']
    2
    >>> b
    1

    """
    return run_compiled_python_code(compile_python_code(code), namespace or {})


def ispartialsubclass(obj:object, cls:object or (object,)) -> bool:
    return ispartial(obj, cls, subclass=True)
def ispartialinstance(obj:object, cls:object or (object,)) -> bool:
    return ispartial(obj, cls, instance=True)

def ispartial(obj:object, cls:object or (object,), subclass:bool=False,
              instance:bool=False) -> bool:
    if obj is cls: return True
    while type(obj) is partial:
        obj = obj.func

    # make it a tuple
    clss = cls if isinstance(cls, tuple) and obj is not tuple else (cls,)

    if subclass and instance:
        raise ValueError("Can't test for both subclass and instance")
    if subclass:
        test = lambda obj, cls: type(obj) is type and issubclass(obj, cls)
    elif instance:
        test = lambda obj, cls: isinstance(obj, cls)
    else:
        test = lambda obj, cls: obj is cls
    return any(test(obj, cls) for cls in clss)


def reverse_dict(d:dict, aggregator=set, multiple_values:bool=False) -> dict:
    """Return a new dict containing keys as values (aggregated into an `aggregator`)
    and values as keys.

    >>> sorted(tuple(reverse_dict({1: 2, 2: 3, 4: 2}).items()))
    [(2, {1, 4}), (3, {2})]

    """
    out = defaultdict(list)
    if multiple_values:
        for key, vals in d.items():
            for val in vals:
                out[val].append(key)
    else:
        for key, val in d.items():
            out[val].append(key)
    return {key: aggregator(val) for key, val in out.items()}



# encoding: utf8
"""Routines manipulating the dot.

"""

import itertools
import pydot
from .asp_to_dot import VisualConfig


DEFAULT_DOT_FILE = 'out/out.dot'
DEFAULT_PROG_FILE = 'dot'


def _dot_from_properties(properties:dict or None, prefix:str=' ') -> str:
    """Return a dot '[]' expression where given properties are represented.

    If given properties are None, empty string will be output.

    """
    if properties:
        content = ' '.join(
            '{}="{}"'.format(field, value.replace('"', r'\"'))
            for field, value in properties.items()
        )
        return prefix + '[' + content + ']'
    else:
        return ''


def multiple_graphs_from_configs(visual_configs:[VisualConfig]) -> [[str]]:
    """Yield generators of lines of dot describing the given VisualConfig instances.

    Produce one graph per VisualConfig.
    See function counterpart, one_graph_from_configs.

    """
    for visual_config in visual_configs:
        yield itertools.chain(
            ['Digraph biseau_graph {\n'],
            _from_config(visual_config),
            ['}\n\n\n'],
        )


def one_graph_from_configs(visual_configs:[VisualConfig]) -> [str]:
    """Yield lines of dot describing the given VisualConfig instances.

    Produce only one graph,
    using the union of visual_configs to implement the view.

    """
    yield 'Digraph biseau_graph {\n'
    for visual_config in visual_configs:
        yield from _from_config(visual_config)
    yield '}'


def _from_config(visual_config:VisualConfig) -> [str]:
    """Yield lines of dot's graph describing the given VisualConfig instance.

    """
    arcs, nodes, properties, upper_annotations, lower_annotations, globals_props, ranks = visual_config
    for object, props in globals_props.items():
        yield '\t{}{};\n'.format(object, _dot_from_properties(props))
    if 'node' not in globals_props:
        yield '\tnode [shape=ellipse style=filled width=.25]\n'
    if 'edge' not in globals_props:
        yield '\tedge [arrowhead=none labeldistance=1.5 minlen=2]\n'
    treated_nodes = set()  # contains nodes already treated
    for node in nodes:
        if node not in treated_nodes:
            treated_nodes.add(node)
            yield from _dot_from_node(node, visual_config)
    for source, target in arcs:
        for node in source, target:
            if node not in treated_nodes:
                treated_nodes.add(node)
                yield from _dot_from_node(node, visual_config)
        yield '\t{}->{}{}\n'.format(source, target, _dot_from_properties(properties.get((source, target))))
    # build ranks
    for ranktype, nodes in ranks.items():
        yield '\t{{rank={}; {}}}\n'.format(ranktype, ';'.join(nodes))


def _dot_from_node(node, visual_config) -> [str]:
    "Yield dot content describing given node in given visual_config"
    arcs, nodes, properties, upper_annotations, lower_annotations, globals_props, ranks = visual_config
    node_dot_props = _dot_from_properties(properties.get(node))
    yield '\t{n}{d}\n'.format(n=node, d=node_dot_props)
    if lower_annotations.get(node):
        yield '\t{n} -> {n} {d}\n'.format(
            n=node, d=_dot_from_properties(lower_annotations[node]))
    if upper_annotations.get(node):
        yield '\t{n} -> {n} {d}\n'.format(
            n=node, d=_dot_from_properties(upper_annotations[node]))


def dot_to_png(dot_lines:iter, outfile:str, dotfile:str=DEFAULT_DOT_FILE,
               prog:str=DEFAULT_PROG_FILE):
    """Write in outfile a png render of the graph described in given dot lines"""
    dot_lines = ''.join(dot_lines)
    if dotfile:
        with open(dotfile, 'w') as fd:
            fd.write(dot_lines)
    graphs = iter(pydot.graph_from_dot_data(dot_lines))
    for graph in graphs:
        with open(outfile, 'wb') as fd:
            fd.write(graph.create(prog=prog or DEFAULT_PROG_FILE, format='png'))
        break

    # TODO: handle multiple graphs (see #10)
    for graph in graphs:
        print('WARNING: an additionnal graph as been found in the dot file.'
              'It will be ignored.')
        print(graph)



# encoding: utf8
"""Biseau script.

Compute concepts and AOC poset as a single model.


"""

import os
import clyngor


NAME = 'Concepts'
TAGS = {'FCA', 'initial example'}
ERASE_CONTEXT = False

ASP_CONCEPTS = {
    'formal': 'data/asp/yield_concepts.lp',
    'object_oriented': 'data/asp/yield_object_oriented_concepts.lp',
    'property_oriented': 'data/asp/yield_property_oriented_concepts.lp',
}
ASP_AOCPOSET = 'data/asp/yield_aocposet.lp'
ASP_AOCPOSET_ONLY = 'data/asp/yield_aocposet_only.lp'
ASP_SUPINFIMUMS = 'data/asp/build_inum.lp'
ASP_NOSYMMETRY = 'data/asp/no-symmetry.lp'


def _output_predicates(type, only_aoc, supremum_and_infimum, remove_symmetry, careful_parsing):
    keys = ('concept', 'specext', 'specint')
    keys += () if only_aoc else ('ext', 'int')
    return frozenset(keys)


def outputs(type, only_aoc, supremum_and_infimum, remove_symmetry, careful_parsing):
    return frozenset(key + ('/1' if key == 'concept' else '/2') for key in
                     _output_predicates(type, only_aoc, supremum_and_infimum, remove_symmetry, careful_parsing))
INPUTS = {'rel/2'}
OUTPUTS = outputs(type='formal', only_aoc=False, supremum_and_infimum=True, remove_symmetry=False, careful_parsing=False)


def run_on(context:str, *,
           type:list=tuple(sorted(tuple(ASP_CONCEPTS.keys()))),
           only_aoc:bool=False,
           supremum_and_infimum:bool=True,
           remove_symmetry:bool=False,
           careful_parsing:bool=True):
    """
    type -- type of concepts to yield
    only_aoc -- do not yield ext/2 and int/2
    supremum_and_infimum -- yield supremum and infimum, even if not in data
    remove_symmetry -- remove concepts with minimal item not in objects sets (has meaning only for symmetric contexts)
    careful_parsing -- use a slower but more robust parser for input atoms
    """
    codes = (
        [ASP_CONCEPTS.get(type, ASP_CONCEPTS['formal']), ASP_AOCPOSET]
        + ([ASP_AOCPOSET_ONLY] if only_aoc else [])
        + ([ASP_NOSYMMETRY] if remove_symmetry else [])
    )
    models = clyngor.solve(
        codes, inline=context, stats=False,
        constants={'allow_non_concept': int(supremum_and_infimum)},
    ).by_predicate
    if careful_parsing:
        models = models.careful_parsing
    keys = _output_predicates(type, only_aoc, supremum_and_infimum, remove_symmetry, careful_parsing)
    for idx, model in enumerate(models):
        yield 'concept({}).\n'.format(idx)
        for key in keys:
            thgs = model.get(key, ())
            for thg in thgs:
                thg, = thg  # first arg only: obj and att have an arity of 1
                yield '{}({},{}). '.format(key, idx, thg)
        yield '\n'



"""Entry point for package.

To play with scripts parameters, you must use the configuration
file instead of the direct enumeration of filenames.

"""

import os
import sys
import argparse
import itertools
import clyngor
from . import core
from . import module_loader
from . import __version__


def parse_cli(args:iter=None) -> dict:
    # main parser
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument('infiles', type=str, nargs='*', metavar='MODULE',
                        default=[], help='files containing ASP or Python code')
    parser.add_argument('--outfile', '-o', type=str, default='out.png',
                        help="output file. Will be overwritten with png data. Can be templated with '{}'")
    parser.add_argument('--dotfile', '-d', type=str, default=None,
                        help="output file. Will be overwritten with dot data. Can be templated with '{}'")
    parser.add_argument('--config', '-c', type=str, default=None,
                        help="configuration file, specifying scripts and their options")
    parser.add_argument('--gif-duration', '-gd', type=int, default=500,
                        help="gif duration in millisecond")
    parser.add_argument('--nb-model', '-n', type=int, default=0,
                        help="number of model of the final context to handle. 0 for all.")
    parser.add_argument('-v', '--verbosity', action='count', default=0)
    parser.add_argument('--version', action='version', version=f'biseau {__version__}')

    # flags
    parser.add_argument('--gif', '-g', action='store_true', default=False,
                        help="Do not merge graphs ; build a gif with frame for each model")
    parser.add_argument('--stdin', action='store_true', default=False,
                        help="Take input in stdin, reading it as clingo output, where each model is a script added to the script pool")

    return parser.parse_args(args)


if __name__ == '__main__':
    args = parse_cli()
    all_infiles = args.infiles
    if args.config:
        all_infiles = itertools.chain(
            all_infiles,
            core.build_pipeline.from_configfile(args.config, verbosity=args.verbosity)
        )
    if os.path.splitext(args.outfile)[1] == '.gif' and args.gif is None:
        args.gif = True  # turn on the gif behavior
    if args.stdin:  # expect clingo output in stdin
        stdin_scripts = []
        for atoms in clyngor.utils.parse_clingo_output(sys.stdin):
            atoms = clyngor.utils.answer_set_to_str(atoms, atom_end='.', atom_sep='\n')
            stdin_scripts.extend(module_loader.build_scripts_from_asp_code(atoms))
        all_infiles = itertools.chain(all_infiles, stdin_scripts)

    if args.gif:
        core.gif_from_filenames(
            all_infiles,
            giffile=args.outfile,
            dotfile_template=args.dotfile,
            duration=args.gif_duration,
            verbosity=args.verbosity,
            nb_model=args.nb_model,
        )
    else:
        if args.outfile and '{}' in args.outfile:
            builder = core.multiple_images_from_filenames
        else:
            builder = core.single_image_from_filenames
        builder(
            all_infiles,
            dotfile=args.dotfile,
            outfile=args.outfile,
            return_image=False,
            verbosity=args.verbosity,
            nb_model=args.nb_model,
        )



# encoding: utf8
"""Allow user to export (some parts of) the ASP program into a file.

"""

import os
import clyngor
from functools import partial


NAME = 'ASP Extractor'
TAGS = {'utils'}
INPUTS = {}
OUTPUTS = {}

DEFAULT_SHOWS = """% Let the #show begin:
#show link/2.


"""
DEFAULT_EXPORT_FILE = 'extraction.lp'
with open(DEFAULT_EXPORT_FILE, 'ab') as fd:
    pass  # make sure that it exists



def run_on(context:str, *, shows:str=DEFAULT_SHOWS, export_file:(open, 'w')=DEFAULT_EXPORT_FILE):
    """
    shows -- ASP lines that will be used to export, but will not be added to the context
    export_file -- name of the file to export to. If it contains an '{}', each model will generate its own file

    """
    try:
        if not export_file: return context
        models = clyngor.solve((), inline=context + '\n' + shows).atoms_as_string.careful_parsing
        istemplate = isinstance(export_file, str) and '{}' in export_file
        if not istemplate:  # empty it before writing it
            with open(export_file, 'w') as fd:
                pass

        for idx, model in enumerate(models):
            fname = export_file.format(idx) if istemplate else export_file
            with open(fname, 'w' if istemplate else 'a') as fd:
                fd.write('\n\n%% MODEL {}:\n'.format(idx))
                fd.write(''.join(atom + '.\n' for atom in model))
    except clyngor.ASPSyntaxError as err:
        raise err
    except clyngor.utils.ASPSyntaxError as err:
        print('This is specifically coming from utils module ; wtf?!')
        raise err
    return context



# encoding: utf8
"""Widget helping user picking a formal context to use.

If the choosen file is in ASP format, the data will be output as-is.

"""

import os
import glob
import tempfile
import tkinter as tk
from tkinter import ttk
from biseau import format_converters


NAME = 'Context Picker'
TAGS = {'FCA', 'ASP', 'initial example'}
INPUTS = {}
OUTPUTS = {'rel/2'}


def run_on(context:str, *, fname:open=None):
    """
    file -- the lp/cxt/slf/csv file to open. Must contains valid ASP or FCA format.
    """
    if not fname: return f'% {NAME}: no file given.\n'
    if not os.path.exists(fname): return f'% {NAME}: invalid file name: {fname}.\n'
    delete = False
    if os.path.splitext(fname)[1] != '.lp':
        with tempfile.NamedTemporaryFile('w', suffix='.lp', delete=False) as fd:
            format_converters.convert(fname, fd.name)
            fname = fd.name
            delete = True
    with open(fname) as fd:
        return fd.read()
    if delete:
        os.unlink(fname)



# encoding: utf8
"""Routines converting various format into ASP.

"""
import os
import csv
import shutil
import tempfile
import itertools
from biseau import core


def format_from_filename(fname:str) -> str or None:
    """Return the format associated with given filename"""
    return os.path.splitext(fname)[1][1:]


def convert(fin:str, fout:str):
    """Convert content found in input file to the format inferred from
    given output filename, where it's then wrote.

    """
    input_format = format_from_filename(fin)
    output_format = format_from_filename(fout)

    if output_format != 'lp':
        raise NotImplementedError("Not in the scope of this project")

    if input_format == output_format:
        shutil.copy(fin, fout)
    elif input_format not in convert_to_lp:
        raise NotImplementedError("Format {} not handled".format(input_format))
    else:
        convert_to_lp[input_format](fin, fout)


def convert_cxt_to_lp(fin:str, fout:str):
    with open(fin) as ifd, open(fout, 'w') as ofd:
        assert next(ifd) == 'B\n', 'expects a B'
        assert next(ifd) == '\n', 'expects empty line'
        nb_obj, nb_att = map(int, (next(ifd), next(ifd)))
        assert next(ifd) == '\n', 'expects empty line'
        objects = tuple(next(ifd).strip() for _ in range(nb_obj))
        attributes = tuple(next(ifd).strip() for _ in range(nb_att))
        for object, properties in zip(objects, ifd):
            intent = itertools.compress(attributes, (char.lower() == 'x'
                                                     for char in properties))
            for prop in intent:
                ofd.write('rel("{}","{}").'.format(object, prop))


def convert_slf_to_lp(fin:str, fout:str):
    with tempfile.NamedTemporaryFile('w', delete=True) as fd:
        core.slf_to_cxt.file_to_file(fin, fd.name)
        return convert_cxt_to_lp(fd.name, fout)


def convert_txt_to_lp(fin:str, fout:str):
    with open(fin) as ifd, open(fout, 'w') as ofd:
        lines = csv.reader(ifd, delimiter='|')
        attributes = tuple(map(str.strip, next(lines)[1:-1]))  # first and last fields are empty
        for object, *props in lines:
            intent = itertools.compress(attributes, (char.strip().lower() == 'x'
                                                     for char in props))
            for prop in intent:
                ofd.write('rel("{}","{}").'.format(object.strip(), prop))


def convert_csv_to_lp(fin:str, fout:str):
    with open(fin) as ifd, open(fout, 'w') as ofd:
        lines = csv.reader(ifd, delimiter=',')
        attributes = tuple(map(str.strip, next(lines)[1:]))  # first field is empty
        for object, *props in lines:
            intent = itertools.compress(attributes, (char.strip().lower() == 'x'
                                                     for char in props))
            for prop in intent:
                ofd.write('rel("{}","{}").'.format(object.strip(), prop))


convert_to_lp = {
    'cxt': convert_cxt_to_lp,
    'slf': convert_slf_to_lp,
    'txt': convert_txt_to_lp,
    'csv': convert_csv_to_lp,
}



# encoding: utf8
""" Compute pattern concepts applied to integer data and using intervals.

"""

import os
import clyngor


NAME = 'Integer Pattern Concepts'
TAGS = {'FCA'}
ERASE_CONTEXT = False
INPUTS = {'rel/3'}
OUTPUTS = {'ext/2', 'int/2'}

ASP_CODE_SEARCH_CONCEPTS = """
object(O):- rel(O,_,_).
condition(C):- rel(_,C,_).
val(X):- rel(_,_,X).

% Choose a subset of objects
{ ext(O): rel(O,_,_) }.

% The intervals for choosen objects.
interval(C,Min,Max):- condition(C) ; Min=#min{V,O: rel(O,C,V), ext(O)}
                                   ; Max=#max{V,O: rel(O,C,V), ext(O)}.

% Object is valid on Condition.
ok(O,C):- rel(O,C,V) ; interval(C,Min,Max) ; Min<=V ; V<=Max.
% Object is valid for all Conditions.
ok(O):- object(O) ; ok(O,C): condition(C).
% Avoid any model that do not include maximal number of objects.
:- not ext(O) ; ok(O).

% Object is non-valid if value is not in the interval.
% ko(O):- rel(O,C,V) ; interval(C,Min,Max) ; V<Min.
% ko(O):- rel(O,C,V) ; interval(C,Min,Max) ; V>Max.
% Avoid any model that do not include maximal number of objects.
% :- object(O) ; not ext(O) ; not ko(O).
"""


def run_on(context:str, *,
           careful_parsing:bool=True):
    """
    careful_parsing -- use a slower but more robust parser for input atoms
    """
    models = clyngor.solve(inline=context + ASP_CODE_SEARCH_CONCEPTS, stats=False).by_predicate
    if careful_parsing:
        models = models.careful_parsing
    for idx, model in enumerate(models):
        # print('MODEL', idx)
        yield 'concept({}).\n'.format(idx)
        extents = model.get('ext', ())
        for extent in extents:
            # print(extent, end='')
            yield 'ext({},{}). '.format(idx, extent[0])
        # print()
        intervals = model.get('interval', ())
        for interval in intervals:
            # print(interval, '', end='')
            if not interval:  # empty interval, no extent
                yield 'int({},interval({},0,100)). '.format(idx, idx)
            else:
                yield 'int({},interval({})). '.format(idx, ','.join(map(str, interval)))
        # print()
        yield '\n'



# encoding: utf8
"""Generalist mappings, allowing user to easily tune the galois lattice visualization.

"""
import textwrap


NAME = 'Show Galois Lattice'
TAGS = {'FCA', 'initial example'}
INPUTS = {'specint/2', 'specext/2', 'concept/1', 'under/2'}
OUTPUTS = {'annot/3', 'link/2', 'obj_property/3'}


def source_view(concept_uid, use_aoc_as_label, arrows:bool, prettify:bool, dpi:int):
    _ = lambda s, c=True: (textwrap.dedent(s.strip()) if c else '')
    return '% SHOW GALOIS LATTICE\n' + '\n\n'.join((
        _("""% Build the lattice.
link(X,Y):- under(X,Y)."""),
        _("""% I want AOC poset to be written.
annot(upper,X,A):- {s}int(X,A).
annot(lower,X,O):- {s}ext(X,O).
""".format(s='spec' if use_aoc_as_label else '')),
        _("""
% Show the concept uid as node label.
label(X,X):- concept(X).
""", concept_uid),
        _("""
% Get a good visualization.
obj_property(graph,dpi,{}).
""".format(dpi), int(str(dpi).strip('"')) > 0),
        _("""
% Delete arrows head.
obj_property(edge,arrowhead,none).
""", not arrows),
        _("""
% Some objects properties to prettify the visualization.
obj_property(edge,labeldistance,\"1.5\").
obj_property(edge,minlen,2).
""", prettify),
    ))


def run_on(context, *,
           concept_uid:bool=True,
           use_aoc_as_label:bool=True,
           arrows:bool=False,
           prettify:bool=True,
           dpi:int=300):
    """

    concept_uid -- show the concept unique id as node label
    use_aoc_as_label -- use the AOC poset to avoid too crowded lattice (do not works well with object/property oriented lattices)
    arrows -- show arrows of edges
    prettify -- set of various settings that improve the lattice feeling of the graph
    dpi -- increase it for sharper visualization

    """
    return source_view(concept_uid, use_aoc_as_label, arrows, prettify, dpi)



# encoding: utf8
"""

Populate existing lattice representation with the data of influency graph.


"""

import os
import clyngor
from collections import defaultdict


NAME = 'Influency graph'
TAGS = {'FCA'}
TAGS = {'FCA', 'initial example'}  # todel
ERASE_CONTEXT = False

INPUTS = {'smaller/2', 'link/2', 'specext/2', 'specint/2'}
OUTPUTS = {'link/2', 'dot_property/3'}

ASP_FILTER_LATTICE_SYMMETRY = """
keep(N) :- concept(N,_) ; Om=#min{O:ext(N,O)} ; Am=#min{A:int(N,A)} ; Am>=Om.
#show.
#show link(A,B): link(A,B), keep(A), keep(B).
#show concept(C): keep(C).
"""  # TODO
ASP_COLORIZE = """
bigger(C1,C2) :- smaller(C2,C1).
color(C,green) :- not bigger(D,C): specint(D,_) ; bigger(_,C).
color(C,red) :- not smaller(D,C): specext(D,_) ; smaller(_,C).
"""
ASP_SHOW_INFLUENCES = """
link(S,T)                          :- influence(S,T,_).
dot_property(S,T,constraint,false) :- influence(S,T,_).
dot_property(S,T,arrowhead,normal) :- influence(S,T,_).
dot_property(S,T,color,red)        :- influence(S,T,_).
dot_property(S,T,fontcolor,red)    :- influence(S,T,_).
dot_property(S,T,label,L)          :- influence(S,T,L).
"""
ASP_AMBIGOUS_CANDIDATE = """
concept(C,obj,N) :- concept(C) ; ext(C,N).
concept(C,att,N) :- concept(C) ; int(C,N).

ambigous_candidate(C1,C2,N,M):-
    C1!=C2
    ; concept(C1,obj,N) ; concept(C1,att,M)
    ; concept(C2,S,N) ; concept(C2,S,M)
    ; not included(C1,C2)
    .
included(C1,C2) :- C1!=C2 ; concept(C2,T,X): concept(C1,U,X) ; concept(C1,U,_) ; concept(C2,T,_).

#show.
#show ambigous_candidate/4.
#show concept/3.
"""


def run_on(context:str, *,
           colorize:bool=True,
           show_influences:bool=True,
           keep_influences_symmetry:bool=False):
    """
    colorize -- show AOC poset min/max concepts in red and green
    show_influences -- render the influence edges in red
    keep_influences_symmetry -- keep only a non-redoundant subset of influencies
    """
    # get influences between concepts
    influences = defaultdict(lambda: (set(), set()))
    concepts = defaultdict(lambda: (set(), set()))
    # print(context)
    models = clyngor.solve(
        inline=context+ASP_AMBIGOUS_CANDIDATE,
        stats=False, constants={'allow_non_concept': 1},
    ).careful_parsing.by_predicate
    # print('CMD:', models.command)
    for model in models:
        for src, trg, ea, eb in model.get('ambigous_candidate', ()):
            influences[src, trg][0].add(ea)
            influences[src, trg][1].add(eb)
        for concept, objset, elem in model.get('concept', ()):
            concepts[concept][0 if objset == 'obj' else 1].add(elem)
    # print('INFLUENCES:', dict(influences))
    # print('  CONCEPTS:', dict(concepts))
    # yield influencies relations
    def ismotif(concept, concepts=concepts) -> bool:
        "True if given concept is a motif"
        obj, att = concepts[concept]
        return min(obj) <= min(att)
    prettyfy = lambda s: '{' + ','.join(sorted(tuple(s))) + '}'
    for (src, trg), (objs, atts) in influences.items():
        if not keep_influences_symmetry and not (ismotif(src) and ismotif(trg)):
            continue
        yield f'influence({src},{trg},"{prettyfy(objs) + "×" + prettyfy(atts)}").'
    if show_influences:
        yield ASP_SHOW_INFLUENCES
    if colorize:
        yield ASP_COLORIZE



# encoding: utf8
"""The core part of visualization, compiling ASP into dot.

"""

import textwrap as textwrap_module
from collections import namedtuple, defaultdict
from biseau import utils


RANK_TYPES = {'same', 'min', 'source', 'max', 'sink'}
DOTABLE_PREDICATES = {'link', 'node', 'color', 'shape', 'label', 'annot', 'dot_property', 'obj_property', 'textwrap', 'rank'}
VisualConfig = namedtuple('VisualConfig', 'arcs, nodes, properties, upper_annotations, lower_annotations, global_properties, ranks')
"""
    arcs -- iterable of 2-uplet (source's uid, target's uid)
    nodes -- set of nodes that may or may not be used by arcs
    properties -- map uid -> (field -> value) and (uid, uid) -> (field -> value)
    upper_annotations -- map uid -> {field -> value} specialized for annotations
    lower_annotations -- map uid -> {field -> value} specialized for annotations
    global_properties -- map object -> (field -> value), with object in (graph, edge, node)
    ranks -- map rank-type -> iterable of sets of node

Properties are mapping directly dot properties to nodes (single uid)
or edges (two uid). This allow user to build very precisely the output dot.

"""


def visual_config_from_asp(asp_models:iter, annotation_sep:str=' ') -> [VisualConfig]:
    """Yield VisualConfig instances initialized according to rules
    found in each given asp models.

    asp_models -- iterable of clyngor.Answers instances
    annotation_sep -- separator between each annotation content on same node

    """
    for model in asp_models.by_predicate:
        base_atoms = defaultdict(set)  # predicate -> {args}
        viz_atoms = defaultdict(lambda: defaultdict(set))  # level -> {predicate: {args}}
        for viz in model.get('viz', ()):
            if len(viz) == 1:
                pred, args = viz[0]
                base_atoms[pred].add(args)
            elif len(viz) == 2:  # the first param is here to define the level
                pred, args = viz[1]
                viz_atoms[int(viz[0])][pred].add(args)
        # put the predicate not in viz() in the base level
        for predicate in DOTABLE_PREDICATES:
            base_atoms[predicate] |= frozenset(model.get(predicate, ()))
        if viz_atoms:
            for level, atoms in viz_atoms.items():
                yield visual_config_from_atoms(atoms, base_atoms, annotation_sep)
        else:  # no viz atoms
            yield visual_config_from_atoms({}, base_atoms, annotation_sep)


def visual_config_from_atoms(atoms:dict, base_atoms:dict,
                             annotation_sep:str) -> VisualConfig:
    arcs, nodes = [], set()
    upper_annotations = defaultdict(lambda: defaultdict(set))
    lower_annotations = defaultdict(lambda: defaultdict(set))
    properties = defaultdict(lambda: defaultdict(set))  # node -> (property -> {value})
    global_properties = defaultdict(lambda: defaultdict(set))  # dot object -> (property -> value)
    ranks = defaultdict(set)  # rank-type -> {node}
    max_label_width = {}  # object: maximal text width
    labeldistance_multiplier = None

    def get_atoms_of_predicate(predicate:str):
        assert predicate in DOTABLE_PREDICATES, predicate
        yield from base_atoms.get(predicate, ())
        yield from atoms.get(predicate, ())

    def get_uid_from_atom(atom:str or tuple):
        if isinstance(atom, str):
            return atom
        elif isinstance(atom, tuple):
            if len(atom) == 2:  # a regular atom
                if len(atom[1]) == 0:
                    return atom[0]
                else:  # atom with args
                    return '{}({})'.format(atom[0], ','.join(map(get_uid_from_atom, atom[1])))
        raise ValueError(f"Malformed node uid of type '{type(atom)}' found: {atom}")

    for link in get_atoms_of_predicate('link'):
        if len(link) == 2:
            arcs.append(tuple(map(get_uid_from_atom, link)))
    for node in get_atoms_of_predicate('node'):
        if len(node) == 1:
            nodes.add(tuple(map(get_uid_from_atom, node))[0])
    for args in get_atoms_of_predicate('textwrap'):
        if len(args) == 1:  # global value
            max_label_width[None] = int(args[0])
        elif len(args) == 2:
            node, value = args
            max_label_width[node] = int(value)
        elif len(args) == 3:
            src, trg, value = args
            max_label_width[src, trg] = int(value)
    for annotation in get_atoms_of_predicate('annot'):
        if len(annotation) == 1:  # determine label distance multiplier
            labeldistance_multiplier = float(annotation[0].strip('"'))
        if len(annotation) == 3:
            type_, node, content = annotation
            node = get_uid_from_atom(node)
            if type_ == 'upper':
                upper_annotations[node]['taillabel'].add(content.strip('"'))
            elif type_ == 'lower':
                lower_annotations[node]['headlabel'].add(content.strip('"'))
            elif type_ == 'label':
                properties[node]['label'].add(content.strip('"'))
            else:
                print('Unknow annotation type: {}'.format(type_))
        elif len(annotation) == 4:  # other field
            type_, node, field, content = annotation
            node = get_uid_from_atom(node)
            if type_ == 'upper':
                upper_annotations[node][field].add(content.strip('"'))
            elif type_ == 'lower':
                lower_annotations[node][field].add(content.strip('"'))
    for property in get_atoms_of_predicate('dot_property'):
        if len(property) == 3:  # it's for node
            node, field, value = property
            node = get_uid_from_atom(node)
            properties[node][field.strip('"')].add(value.strip('"'))
        elif len(property) == 4:  # it's for edges
            src, trg, field, value = property
            src, trg = map(get_uid_from_atom, (src, trg))
            properties[src, trg][field.strip('"')].add(value.strip('"'))
    for ranking in get_atoms_of_predicate('rank'):
        if len(ranking) == 2:  # rank, node
            ranktype, node = ranking
            if ranktype not in RANK_TYPES:
                print("WARNING: atom rank({},{}) describe a rank with unknow "
                      "type {}. Expected types: {}."
                      "".format(*ranking, ranktype, ', '.join(RANK_TYPES)))
            ranks[ranktype].add(node)
    for colored in get_atoms_of_predicate('color'):
        if len(colored) == 2:  # node
            node, color = colored
            node = get_uid_from_atom(node)
            properties[node]['fillcolor'].add(color)
        elif len(colored) == 3:  # edge
            src, trg, color = colored
            src, trg = map(get_uid_from_atom, (src, trg))
            properties[src, trg]['color'].add(color)  # fillcolor do not exists for edges
    for shaped in get_atoms_of_predicate('shape'):
        if len(shaped) == 2:  # node
            node, shape = shaped
            node = get_uid_from_atom(node)
            properties[node]['shape'].add(shape)
    for labeled in get_atoms_of_predicate('label'):
        if len(labeled) == 2:  # node
            node, label = labeled
            node = get_uid_from_atom(node)
            properties[node]['label'].add(label)
        elif len(labeled) == 3:  # edge
            src, trg, label = labeled
            src, trg = map(get_uid_from_atom, (src, trg))
            properties[src, trg]['label'].add(label)
    for property in get_atoms_of_predicate('obj_property'):
        if len(property) == 3:
            obj, field, value = map(lambda s:str.strip(s, '"'), property)
            if obj not in {'graph', 'edge', 'node'}:
                print('WARNING: object property {} is unexpected, and may '
                      'lead to error in generation.'.format(obj))
            if field in global_properties:
                print('WARNING: object property {} set multiple times with {} replacing {}.'
                      ''.format(field, value, graph_properties[field]))
            global_properties[obj.strip('"')][field] = value

    # posttreat the data for later use
    arcs = tuple(arcs)
    def treat_texts(texts:iter, node, max_label_width=max_label_width) -> str:
        ret = annotation_sep.join(map(get_uid_from_atom, texts)).strip('"')
        text_width = max_label_width.get(node, max_label_width.get(None))
        if text_width:
            ret = textwrap_module.fill(ret, width=int(text_width))
        return ret

    for node in upper_annotations:
        props = upper_annotations[node]
        props.setdefault('color', 'transparent')
        props.setdefault('labelangle', '90')
        for key in tuple(props):
            if not isinstance(props[key], str):
                text = treat_texts(props[key], (node, node))
                props[key] = text
            if labeldistance_multiplier is not None:
                # print('TEXT:', repr(text), len(text.split('\\n')))
                props['labeldistance'] = str(1 + labeldistance_multiplier * (len(text.split('\\n')) - 1))

    for node in lower_annotations:
        props = lower_annotations[node]
        props.setdefault('color', 'transparent')
        props.setdefault('labelangle', '270')
        for key in tuple(props):
            if not isinstance(props[key], str):
                text = treat_texts(props[key], (node, node))
                props[key] = text
            if labeldistance_multiplier is not None:
                # print('TEXT:', repr(text), len(text.split('\\n')))
                props['labeldistance'] = str(1 + labeldistance_multiplier * (len(text.split('\\n')) - 1))

    for key in properties:
        if 'color' in properties[key]:
            try:
                properties[key]['color'] = utils.color_from_colors(properties[key]['color'])
            except ValueError:  # invalid color
                properties[key]['color'] = next(iter(properties[key]['color']))
        if 'fillcolor' in properties[key]:
            try:
                properties[key]['fillcolor'] = utils.color_from_colors(properties[key]['fillcolor'])
            except ValueError:  # invalid color
                properties[key]['fillcolor'] = next(iter(properties[key]['fillcolor']))
        for field in properties[key]:
            if field not in {'color', 'fillcolor'}:
                properties[key][field] = treat_texts(properties[key][field], key)

    return VisualConfig(
        arcs, frozenset(nodes), dict(properties), dict(upper_annotations), dict(lower_annotations),
        dict(global_properties), ranks
    )



# encoding: utf8
"""Load and validate a python module received as a filename.

"""
import os
import re
import glob
import json
import inspect
import textwrap
import importlib
import itertools
from functools import partial
from collections import defaultdict

import clyngor
from biseau import utils
from biseau import script as script_functions
from biseau.script import Script, Module


DEFAULT_DOC = 'NO SCRIPT DOC PROVIDED.\nFix this by writing a module documentation inside script definition.'
RETURNS_TYPES = {iter, str}
OPTIONS_TYPES = {int, float, bool, str, open, (open, 'r'), (open, 'w')}
TYPE_DEFAULT = {int: 0, float: 0., bool: False, str: '', open: None}
REGEX_OPTION_DESC = re.compile(r'([a-zA-Z0-9_]+)\s*--\s*(.+)$')

class ScriptError(ValueError):
    pass


def gen_files_in_dir(dirname:str, extensions:[str]=('py', 'lp', 'json'),
                       filter_prefixes:[str]='_') -> (str, str):
    "Yield candidate scripts in given dir, based on file extension"
    yield from (
        fname
        for fname in map(os.path.basename, glob.glob('{}/*.{{{}}}'.format(dirname, ','.join(extensions))))
        if not filter_prefixes or (filer_prefixes and not fname.startswith(filter_prefixes))
    )


def build_scripts_from_dir(dirname:str='scripts', options:dict={}) -> [Script]:
    "Yield all scripts found in given directory (not recursive)"
    for file in gen_files_in_dir(dirname):
        yield from build_scripts_from_file(file, options)

def build_scripts_from_file(fname:str, options:dict={}) -> [Script]:
    "Yield all scripts found in given file (note that only JSON files can define multiple scripts)"
    name, ext = os.path.splitext(fname)
    if ext == '.json':
        yield from build_scripts_from_json_file(fname)
    elif ext == '.py':
        try:
            script = build_python_script_from_name(name)
            yield script
        except ScriptError as err:
            print('SCRIPT ERROR:', str(err))
    elif ext == '.lp':
        yield build_asp_script_from_name(fname)
    elif fname.upper().startswith('ASP'):
        if isinstance(options, (list, tuple, str)):
            yield from build_scripts_from_asp_code(options)
        else:
            print(f"Unknow type for ASP code: {type(options)}")
    else:
        print(f"WARNING file '{fname}' was not recognized")


def merge_scripts_lists(*scripts_lists:iter) -> iter:
    """Yield scripts, ordered according to their dependancies"""
    yield from sort_scripts_per_dependancies(itertools.chain.from_iterable(scripts_lists))


def sort_scripts_per_dependancies(scripts:iter) -> iter:
    """Topological sort of scripts based on their inputs/outputs.

    Do not handle scripts interdependancies.

    """
    scripts = tuple(scripts)
    inputs = {script: frozenset(script.inputs) for script in scripts}
    outputs = {script: frozenset(script.outputs) for script in scripts}
    yield from topological_sort_by_io(inputs, outputs)


def topological_sort_by_io(inputs:dict, outputs:dict) -> iter:
    """Yield keys of inputs and outputs so that a value yielded after another
    is either in need of the previous's outputs, or unrelated.

    inputs -- mapping {value: {input}}
    outputs -- mapping {value: {output}}

    """
    # decide {pred: {succs}} for scripts
    topology = defaultdict(set)
    for script, input in inputs.items():
        topology[script]  # just ensure there is one
        for maybe_pred, output in outputs.items():
            if input & output:
                topology[maybe_pred].add(script)
    successors = frozenset(itertools.chain.from_iterable(topology.values()))
    sources = {script for script in topology if script not in successors}
    # compute source, and decide a path
    prev_len = None
    while topology:  # while catch cycles
        while len(topology) != prev_len:
            prev_len = len(topology)
            yield from sources
            topology = {script: {succ for succ in succs if succ not in sources}
                        for script, succs in topology.items()
                        if script not in sources}
            successors = frozenset(itertools.chain.from_iterable(topology.values()))
            sources = {script for script in topology if script not in successors}
        if topology:  # there is at least one cycle
            # take a predecessor, say it is a source
            forced_source = next(iter(topology.keys()))
            sources = {forced_source}
            prev_len = None


def build_python_script_from_name(module_name) -> Script:
    path = module_name.replace('/', '.')
    import_as_module = True
    try:
        module = importlib.import_module(path)
    except TypeError:
        import_as_module = False

    if import_as_module:
        # Reload needed because the module itself is
        #  modified by build_script_from_module
        module = importlib.reload(module)
        return build_script_from_module(module)
    else:  # just load it savagely
        # module_name is not importable directly
        return build_script_from_json({'python file': module_name + '.py'})

def build_python_script_from_name(module_name) -> Script:
    "Replace the previous implementation. Limited, but we don't have to manage python modules as specific objects"
    return build_script_from_json({'python file': module_name + '.py'})


def build_asp_script_from_name(fname:str) -> str:
    with open(fname) as fd:
        asp_code = fd.read()
    name = os.path.splitext(os.path.basename(fname))[0]
    name.replace('_', ' ')
    with open(fname) as fd:
        description = []
        for line in fd:
            if line.startswith('% '):
                description.append(line[2:])
            else: break
    description = '\n'.join(description)
    # reuse the json interface
    return build_script_from_json({
        'name': name,
        'ASP': asp_code,
        'description': description,
        'inputs': [],
        'outputs': [],  # TODO: search for #show's in the file
    })


def build_scripts_from_json_file(fname:str) -> [Script]:
    """Yield Script instances found in given file in JSON format"""
    with open(fname) as fd:
        data = json.load(fd)
    if isinstance(data, list):  # multiple scripts
        for payload in data:
            yield build_script_from_json(payload)
    elif isinstance(data, dict):  # only one
        yield build_script_from_json(data)
    else:
        raise ScriptError("Given json file {} is not correctly formatted. "
                          "First object should be a list or a dict, not a {}"
                          "".format(fname, type(data)))


def build_scripts_from_asp_code(data:str or list) -> [Script]:
    """Yield one Script instance initialized with given source code"""
    if isinstance(data, (tuple, list)):  # multiple scripts
        for source in data:
            yield from build_scripts_from_asp_code(source)
    elif isinstance(data, str):  # only one
        yield build_script_from_json({
            'name': 'inline ASP code',
            'ASP': data,
            'language': 'asp',
            'description': 'inline ASP code',
            'inputs': [],
            'outputs': [],  # TODO: search for #show's in the file
        })
    else:
        raise ScriptError(f"Given ASP source of type {type(data)} can't be handled.")


def build_script_from_json(module_def:dict) -> Script:
    """From given JSON build a Script instance"""
    module = Module()

    # I/O
    module.INPUTS = frozenset(module_def.get('inputs', ()))
    module.OUTPUTS = frozenset(module_def.get('outputs', ()))

    # Fields
    module.NAME = module_def.get('name', 'unamed script')
    if 'tags' in module_def: module.TAGS = frozenset(module_def['tags'])
    module.__doc__ = module_def.get('description', DEFAULT_DOC)

    # building the run_on function
    if 'ASP file' in module_def:
        module.language = 'asp file'
        module.source_code = module_def['ASP file']
    elif 'ASP' in module_def:
        module.language = 'asp'
        module.source_code = module_def['ASP']
    elif 'python file' in module_def:
        module.language = 'python file'
        module.source_code = module_def['python file']
    elif 'python' in module_def:
        module.language = 'python'
        module.source_code = module_def['python']
    else:
        raise ValueError(f"JSON script {module.NAME} do not have any code field ('ASP' "
                         "or 'ASP file' for instance). If this script was "
                         "generated with Biseau, it's possible that you're "
                         "using an older version than the script creator."
                         "")
    return build_script_from_module(module)


def build_script_from_module(module, *, defaults:dict={}) -> Script or ScriptError:
    """Low level function. Expect module to be a python module, or a namespace
    emulating one.

    Will try hard to invalidate given module. If it seems valid, return
    a Script instance describing and referencing the module.

    defaults -- mapping from (name, tags, erase_context) to a
                default value to use if the module does not provide it.

    """
    if not hasattr(module, '__doc__'):
        bad_script_error(module, "Docstring (description) is missing")


    if hasattr(module, 'run_on'):
        run_on_func = module.run_on
        args = inspect.getfullargspec(module.run_on)
        # print('\nSCRIPT ARGS:', getattr(module, 'NAME', 'unamed'), args)

        # Return type
        if inspect.isgeneratorfunction(module.run_on):
            pass
        elif inspect.isfunction(module.run_on) and args.annotations.get('return', str) == str:
            pass
        else:
            bad_script_error(module, "run_on object must be a generator of string"
                             " or a function returning a string, not a {}"
                             "".format(type(module.run_on)))

        # Input mode
        first_arg = args.args[0]
        if first_arg == 'context':
            input_mode = str
        elif first_arg == 'models':
            input_mode = iter
        else:
            bad_script_error(module, "run_on first (and only) positional argument"
                             f" must be either 'context' or 'models', not a {first_arg}")

        if len(args.args) != 1:
            # TODO: allows to use regular arguments instead of keyword only
            bad_script_error(module, f"run_on must have only one positional "
                             f"argument, not {len(args.args)}")

        # detect options
        options = []  # list of (arg name, arg type, default, description)
        for arg in args.kwonlyargs:
            argtype = args.annotations.get(arg)
            isgroup = isinstance(argtype, (tuple, list, set, frozenset))
            if not isgroup and argtype not in OPTIONS_TYPES:
                bad_script_error(module, "Option {} does not have a valid annotation "
                                 "({}). Only tuples, lists, (frozen)sets, and primitive types such as {} are accepted"
                                 "".format(arg, argtype, ', '.join(map(str, OPTIONS_TYPES))))
            if isgroup:  # pick an element or a subset of elements as default
                default = argtype[0] if isinstance(argtype, (tuple, list)) else frozenset()
            else:
                default = args.kwonlydefaults.get(arg, TYPE_DEFAULT.get(argtype))
            options.append((arg, argtype, default))
        default_options = {arg: default for arg, _, default in options}

        # add the descriptions to options
        options_descriptions = options_description_from_module(
            module, frozenset(default_options.keys()))
        options = tuple((arg, type, default, options_descriptions.get(arg, ''))
                        for arg, type, default in options)
        # TODO: detect non keyword only parameters, and check their validity.

    else:  # no run_on function… the source code and the language will be enough, we hope…
        options, default_options = (), {}
        input_mode = str
        run_on_func = None

    # source code data
    source_code = getattr(module, 'source_code', None)
    language = getattr(module, 'language', 'python')

    # detect trivias
    tags = frozenset(getattr(module, 'TAGS', defaults.get('tags', {'undefined'})))
    doc = DEFAULT_DOC
    if module.__doc__:
        doc = '\n'.join(textwrap.wrap(textwrap.dedent(module.__doc__).strip()))

    # build and return the Script instance
    return Script(
        name=getattr(module, 'NAME', defaults.get('name', 'unamed module')),
        description=doc,
        tags=tags,
        module=module,
        options=tuple(options),
        options_values={},
        input_mode=input_mode,
        incompatible=frozenset(getattr(module, 'INCOMPATIBLE', ())),
        **_build_and_validate_io(module, default_options),
        source_code=source_code,
        language=language,
        erase_context=bool(getattr(module, 'ERASE_CONTEXT', defaults.get('erase_context', False))),
        run_on=run_on_func,
    )


def bad_script_error(script, msg:str):
    """Helper to raise errors while building a script"""
    raise ScriptError("Module {} is not a valid script. {}."
                      ''.format(script, msg))


def options_description_from_module(module, options, regex=REGEX_OPTION_DESC) -> dict:
    """Return found description for given options in module"""
    if not module.run_on.__doc__: return {}
    ret = {}  # option: description
    lines = module.run_on.__doc__.splitlines(False)
    for line in lines:
        match = regex.fullmatch(line.strip())
        if match:
            name, desc = match.groups()
            if name in options:
                ret[name] = desc.strip()
    return ret


def _build_and_validate_io(module, default_options:dict={}) -> {str: callable}:
    """Return spec_inputs, spec_outputs, inputs and outputs functions
    built from given module.

    module -- the module containing the things
    default_options -- the options to send to inputs and outputs functions

    return -- the dict {field name: function}, usable directly to create
    a Script instance.

    """
    fields = {}  # field name: field value

    IN = lambda *_, **__: frozenset(getattr(module, 'INPUTS', ()))
    OUT = lambda *_, **__: frozenset(getattr(module, 'OUTPUTS', ()))
    fields['spec_inputs'] = getattr(module, 'inputs', IN)
    fields['spec_outputs'] = getattr(module, 'outputs', OUT)
    fields['inputs'] = IN
    fields['outputs'] = OUT


    # Verify that their are functions, and well dev
    for field in ('spec_inputs', 'spec_outputs', 'inputs', 'outputs'):
        func = fields[field]
        if not callable(func):
            bad_script_error(module, 'Attribute {} is not a function'.format(func))
        if not field.startswith('spec_'):  # it's a class method, not an instance one
            retvalue = func()
            if not isinstance(retvalue, (set, frozenset)):
                bad_script_error(module, "Function {} should return a (frozen)set, "
                                 "not {}".format(func.__name__, type(retvalue)))
    return fields



"""Definitions and functions applied to Script object.

The Script object waits for many parameters, detailed in its constructor,
and its main interface is the run_on method.

The run_on method, constituting the main API interface, IS A NIGHTMARE to generate.
From the Script point of view, the inner _run_on method (gotten from the module,
like JSON), maps a context/models and options to a string/generator of string.
The call_script function is uniformizing all of these in a single interface.

When the source_code attribute of a Script is modified, this run a machinery
whose goal is to build the _run_on method.
This machinery is not runned if run_on is given in the Script constructor.
The machinery is basically an extraction of the source code depending
of the language (python, asp, python file or asp file) and generate a standard
_run_on method based on that.
But because the script can also specify a (new) name, a (new) value for erase_context
and other attributes, that are parsed in module_loader.build_script_from_module,
it is necessary to call that function on the new module (induced by a source_code change).
Therefore, a new Script instance is built, from which all attributes are copied in the current instance.
The code is on the verge to lock itself in an infinite loop.

I'm sorry for the ones coming after.


"""
import clyngor
import traceback
from biseau import utils


class Script:

    def __init__(self, name:str, tags:set, description:str, module, options, options_values, input_mode, incompatible, spec_inputs, spec_outputs, inputs, outputs, source_code, language, erase_context, *, run_on:callable or None=None):
        self.name = str(name)  # human readable name
        self.tags = set(map(str, tags))  # set of tags identifying the script
        self.description = str(description)  # human readable and high level description of the script
        self.module = module  # reference to the module itself
        self.options = tuple(options)  # list of (name, type, default, description) describing each option
        self.options_values = dict(options_values)  # mutable mapping allowing to set options value to be used
        self.input_mode = input_mode  # define if run_on must receive the context or the resulting ASP models
        self.incompatible = set(incompatible)  # list of incompatibles modules
        self.spec_inputs = spec_inputs  # function in module to call to get the inputs knowing the parameters
        self.spec_outputs = spec_outputs  # function in module to call to get the outputs knowing the parameters
        self.inputs = inputs  # function in module to call to get all possible inputs
        self.outputs = outputs  # function in module to call to get all possible outputs
        self.language = str(language)  # string indicating the language implementing the source code
        self.erase_context = bool(erase_context)  # true if the script erase the context (default: false, context is kept)

        # creation of run_on function
        self.__source_code = str(source_code) if source_code else None  # don't trigger the setter
        assert source_code != 'None', "this already happened, it was because previous line called str() on source_code, which could be None"
        self._run_on = run_on
        if not self._run_on:
            self._run_on = compile_run_on(self.language, self.__source_code, self._update_from_module)
        else:
            self._run_on._source = self.__source_code

    def run_on(self, context:str, *args, **kwargs):
        return call_script(self, context, *args, **kwargs)

    @property
    def source_code(self) -> str:
        return self.__source_code

    @source_code.setter
    def source_code(self, new_source:str):
        self.__source_code = str(new_source)
        if self._run_on._source != self.source_code:
            self._run_on = compile_run_on(self.language, self.source_code, self._update_from_module)

    def _update_from_module(self, module):
        """Change all self attributes based on those of given module,
        especially language and source_code defining run_on

        This shouldn't be called when run_on is provided.

        """
        from biseau.module_loader import build_script_from_module
        script = build_script_from_module(module, defaults=vars(self))
        for key in vars(self):
            if getattr(script, key, None) is not None:
                setattr(self, key, getattr(script, key))
        if self._run_on:
            self._run_on._source = self.source_code
        # print('\tFROM:', {k: v for k, v in vars(module).items() if not k.startswith('_')})
        # print('\t TO :', {k: v for k, v in vars(self).items() if not k.startswith('_')})

    @staticmethod
    def call(script, context:str, *args, **kwargs):
        return call_script(script, context, *args, **kwargs)

    def __str__(self):
        return f"<Script language={self.language} code={bool(self.source_code)} #lines={len((self.source_code or '').splitlines())} erase={self.erase_context}>"


class Module:
    """Placeholder for a module/namespace.

    It's assumed safe to make them hashable on content. Also, the hashable
    property is only used during validation and initial core treatments.

    """
    def __hash__(self):
        return hash(tuple(self.__dict__.values()))

    @staticmethod
    def from_dict(mapping:dict) -> object:
        "Create Module instance with given dictionnary as members"
        module = Module()
        for key, val in mapping.items():
            setattr(module, key, val)
        return module



def solve_context(context:str, *, nb_model:int=0) -> clyngor.Answers:
    """Uniformized way to solve an ASP context"""
    return clyngor.solve(inline=context, nb_model=nb_model, options='--project').by_predicate.careful_parsing.int_not_parsed


def call_script(script, context:str, verbosity:int=0):
    """Uniformized call to a Script object"""
    if script.input_mode is str:
        input_data = context
    else:  # need a solving
        assert script.input_mode is iter
        input_data = solve_context(context)
    if verbosity >= 2:  print('RUN run_on… ', end='', flush=True)
    new_context = utils.join_on_genstr(script._run_on)(input_data, **script.options_values)
    if verbosity >= 2:  print('OK!')
    if verbosity >= 3:  print('NEW CONTEXT:', new_context)
    return new_context + ('\n' if script.erase_context else ('\n' + context))


def build_module_from_python_code(pycode:str, **module_options:dict):
    """Low level function, expecting some python code in string, and returning a namespace/module.
    """
    try:
        env = {}
        exec(pycode, env)
    except:
        print(f'Imported Python error:\n', traceback.format_exc())
    # env.update(module_options)
    env = {**module_options, **env}
    env['source_code'] = pycode
    env['language'] = 'python'
    # print('OPTIONS:', module_options, {k: v for k, v in env.items() if not k.startswith('_')})
    module = Module.from_dict({k: v for k, v in env.items() if not k.startswith('__')})
    # print('MODULE:', {k: v for k, v in vars(module).items() if not k.startswith('_')})
    return module


def compile_run_on(language:str, code:str, updater:callable=None) -> callable:
    """Return a runner on given language/code, call given updater with a module
    when a new python module is found

    """
    module = None
    if language == 'python':
        # we need to compile the source code to retrieve the run_on function
        module = build_module_from_python_code(code)
        if not hasattr(module, 'run_on'):  # make a phony run_on
            module.run_on = lambda context: ''
            # print('MODULE w/o run_on:', type(module), {k: v for k, v in vars(module).items() if not k.startswith('__')})
        run_on = module.run_on
    elif language == 'python file':
        # initialize
        with open(code) as fd:
            source = fd.read()
        module = build_module_from_python_code(source)
        if not hasattr(module, 'run_on'):  # make a phony run_on
            module.run_on = lambda context: ''
            # print('MODULE w/o run_on:', type(module), {k: v for k, v in vars(module).items() if not k.startswith('__')})
        run_on = module.run_on
    elif language == 'asp':
        def run_on(context:str):
            return code
    elif language == 'asp file':
        def run_on(context:str):
            with open(code) as fd:
                return fd.read()
    else:
        raise NotImplementedError(f"Language {language} is not supported.")

    run_on._source = code
    if updater and module:  updater(module)
    return run_on



# encoding: utf8
"""Load lp file knowing its path.

"""
import tkinter as tk
import tkinter.filedialog
from functools import partial

NAME = 'Open lp file'
TAGS = {'ASP', 'utils'}


INPUTS = {}
OUTPUTS = {}  # in fact, we do not know. Sorry.

OPEN_LP_FILE = partial(tk.filedialog.askopenfilename, defaultextension='.lp')


def run_on(context:str, *, file:OPEN_LP_FILE=None):
    """
    file -- the lp file to open. Must contains valid ASP.
    """
    if file:
        with open(file) as fd:
            yield fd.read()



__version__ = '0.0.10'

from .script import Script
from .core import run, single_image_from_filenames, multiple_images_from_filenames, gif_from_filenames, compile_to_single_image, compile_to_images, compile_context_to_dots, compile_context_to_dot
from .module_loader import build_scripts_from_file, build_scripts_from_dir
from .export import standalone_export_pipeline



# encoding: utf8
"""Biseau script.

Compute concepts as a single model.

"""

import os
import clyngor


NAME = 'Formal Concepts (simple, no options)'
TAGS = {'FCA'}
INPUTS = {'rel/2'}
OUTPUTS = {'ext/2', 'int/2'}
ASP_CONCEPTS_SRC = 'data/asp/yield_concepts_simple.lp'


def run_on(context:str):
    """Get current ASP encoded context, yield parts of the new context.
    """
    models = clyngor.solve(ASP_CONCEPTS_SRC, inline=context, stats=False).by_predicate
    for idx, model in enumerate(models):
        yield 'concept({}).\n'.format(idx)
        for key in ('ext', 'int'):
            for arg in model.get(key, ()):
                yield '{}({},{}). '.format(key, idx, arg[0])
        yield '\n'



# encoding: utf8
"""Experiments about the scripting API.

"""
import tkinter as tk
from functools import partial

NAME = 'from __future__ import ideas'
DISABLED = True


INPUTS = {}
OUTPUTS = {'user/2'}

LANGUAGES = 'python', 'asp'
OPEN_LP_FILE = partial(tk.filedialog.askopenfilename, default_extension='.lp')
DEFAULT_FILE = '~/programs/biseau/contexts/boolean_3var.lp'

def run_on(context:str, *, coucou_les_copaings:bool=False,
           vas_y_ouvre_moi:OPEN_LP_FILE=DEFAULT_FILE):
    yield context
    if vas_y_ouvre_moi:
        with open(vas_y_ouvre_moi) as fd:
            yield fd.read()
    # yield vas_y_ouvre_moi.read()
    if coucou_les_copaings:
        yield 'rel("copaing","human").'



"""Routines for the export of scripts to various form

"""

import itertools
from . import Script, __version__


def get_pipeline_options(scripts:[Script]) -> dict:
    """Return the mapping option->False to be used by standalone_export_pipeline.

    One may want to set some options to True, to get them in pipeline standalone export.

    """
    return {n: False for _, __, n in option_names_from_options(scripts)}


def standalone_export_pipeline(scripts:[Script], options:dict={}, default_context:str='', verbosity:int=0) -> [str]:
    """Yield Python code strings, implementing a standalone program reproducing given pipeline.

    options -- {option name: bool} indicating whether or not the option
               must be exposed as program option.
    default_context -- the default initial context.
    verbosity -- verbosity of the standalone program itself.

    """
    option_names = {(idx, name): final_name for idx, (_, name, final_name) in enumerate(option_names_from_options(scripts), 1)}
    options_used = tuple(
        (idx, final_name) for idx, final_name in option_names.items()
        if options[final_name]
    )
    yield '"Standalone script generated by biseau ' + str(__version__) + '"\n'
    yield 'import argparse'
    yield 'import clitogui'
    yield 'import clyngor'
    yield ''
    if options:
        yield '@clitogui'
        yield 'def cli():'
        yield '    parser = argparse.ArgumentParser()'
        for (idx, name), final_name in options_used:
            option = scripts[idx-1].options
            option = scripts[idx].options[name]
            print(option)
            yield f'    parser.addArgument({final_name})'
        yield '    ...'
        yield '    return parser'
        yield ''
    yield 'def run_on(context:str, ' + ', '.join(name for _, name in options_used) + '):'
    for script in scripts:
        if script.language == 'python':
            yield '     models = clyngor.solve(inline=context).by_predicate'
        elif script.language == 'asp':
            yield '     context += """\n' + script.source_code + '\n"""'
        else:
            raise ValueError(f"unhandled export of language '{script.language}'")
    yield 'if __name__ == "__main__":'
    yield '    args = cli()'
    yield '    for dot in run_on(**args):'
    yield '        print(dot)'


def option_names_from_options(scripts:[Script]) -> [(str, str, str)]:
    """Yield (script name, option name, unambiguous option name)
    so that unambiguous option name is deterministic and unique
    accross all options of given scripts.

    """
    used_names = set()  # set of all available options
    conflicting_names = set()  # names that will need the script name
    script_names = tuple(script.name for script in scripts)
    script_name_doublons = {n: itertools.count(1) for n in script_names if script_names.count(n) > 1}
    # define names of scripts (append their index when multiple scripts of same name)
    script_names = tuple(
        (script.name + ' ' + str(next(script_name_doublons[script.name])))
        if script.name in script_name_doublons else script.name
        for script in scripts
    )
    assert len(script_names) == len(scripts)
    # detect the name conflicts among scripts options
    for script in scripts:
        for name, type, default, description in script.options:
            if name in used_names:
                conflicting_names.add(name)
            used_names.add(name)
    # for each script, ensure the use of proper option name
    for script_name, script in zip(script_names, scripts):
        for name, type, default, description in script.options:
            final_name = name
            if name in conflicting_names:
                final_name = f'{script_name}: {name}'
            yield script_name, name, final_name



# encoding: utf8
"""Build frequent concepts, allowing one to generate the iceberg lattice.


"""
import clyngor
import tkinter as tk
from functools import partial

NAME = 'Frequent concepts'
TAGS = {'FCA', 'Iceberg Lattice'}
INPUTS = {'rel/2'}
OUTPUTS = {'ext/1', 'int/1'}

ASP_SRC_CONCEPTS = 'data/asp/yield_frequent_concepts.lp'
RATIO_FIELD = partial(tk.Scale, from_=0, to=1, resolution=0.001, orient=tk.HORIZONTAL)


def run_on(models:str, *,
           minsupp_obj:RATIO_FIELD=0.85,
           minsupp_att:RATIO_FIELD=0.):
    """

    minsupp_obj -- minimal support for objects, defining frequent concepts.
    minsupp_att -- minimal support for attributes, defining frequent concepts.

    """
    # print('EXTRACT CONTEXT DATA…')
    model = next(models.by_predicate)
    # print(model)
    assert model.get('allow_non_concept') is None
    if model.get('rel') is None:
        return ''

    # compute minimal number of objects/attributes in the concepts
    min_nb_obj, min_nb_att = 0, 0
    if minsupp_obj > 0:
        nb_obj = len(frozenset(obj for obj, _ in model.get('rel')))
        min_nb_obj = int(float(minsupp_obj) * int(nb_obj))
    if minsupp_att > 0:
        nb_att = len(frozenset(att for _, att in model.get('rel')))
        min_nb_att = int(float(minsupp_att) * int(nb_att))

    # print('COMPUTING FREQUENT CONCEPTS…')
    inlined = clyngor.utils.answer_set_to_str(model, atom_sep='.') + '.'
    models = clyngor.solve(ASP_SRC_CONCEPTS, inline=inlined, constants={
        'minsupp_obj': min_nb_obj,
        'minsupp_att': min_nb_att,
    }).by_predicate

    for idx, model in enumerate(models):
        yield 'concept({}).\n'.format(idx)
        for key in ('ext', 'int', 'specext', 'specint'):
            thgs = model.get(key, ())
            for thg in thgs:
                thg, = thg  # first arg only: obj and att have an arity of 1
                yield '{}({},{}). '.format(key, idx, thg)
        yield '\n'
